-----------------------------------------------------------------------------
--
-- data
--   \healingDone
--     \<healerName>
--       -count, amount, overhealAmount
--       \healDetails
--         \<healName>
--            -count, amount, overhealAmount
--			  \hitDetails
--			    \<healType> -- one of Heal, Critical Heal, Tick
--                -count, amount, overhealAmount
--   \healingTaken
--     \<healTargetName>
--       -count, amount, overhealAmount
--       \healDetails
--         \<healName>
--            -count, amount, overhealAmount
--			  \hitDetails
--			    \<healType> -- one of Heal, Critical Heal, Tick
--                -count, amount, overhealAmount
--
-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentHealing")
L:RegisterTranslations("enUS", function() return {
	["Heal"] = true,
	["Overheal"] = true,
	["Healers List"] = true,
	["Healing Done Details"] = true,
	["Healing Done"] = true,
	["Fight Healing Done"] = true,
	["HPS"] = true,
	["Raw Healing Done"] = true,
	["Raw Healing Done Details"] = true,
	["Raw Healing Done Hit Details"] = true,
	["Fight Raw Healing Done"] = true,
	["Heal/Overheal"] = true,
	["Healing/Overhealing"] = true,
	["Healing Done Hit Details"] = true,

	["Critical Heal Rate"] = true,
	["Critical Heal Rate Details"] = true,

	["Healing Taken"] = true,
	["Fight Healing Taken"] = true,

	["Healing"] = true,
	["Healing Options."] = true,
	["Colors"] = true,
	["Color options"] = true,
	["Healing"] = true,
	["Overhealing"] = true,

	["Tick"] = true,
	["Critical Heal"] = true,
	["Heal"] = true,

	["Heal Data"] = true,
	["Information about each heal used."] = true,
	["Hit Data"] = true,
	["Detailed information about each attack used."] = true,
} end)

L:RegisterTranslations("koKR", function() return {
	["Heal"] = "치유",
	["Overheal"] = "초과 치유",
	["Healers List"] = "힐러 목록",
	["Healing Done Details"] = "치유량 상세정보",
	["Healing Done"] = "치유량",
	["Fight Healing Done"] = "최근 치유량",
	["HPS"] = "HPS",
	["Raw Healing Done"] = "순수 치유량",
	["Raw Healing Done Details"] = "순수 치유량 상세정보",
	["Raw Healing Done Hit Details"] = "순수 치유량 적중 상세정보",
	["Fight Raw Healing Done"] = "최근 순수 치유량",
	["Heal/Overheal"] = "치유/초과치유",
	["Healing/Overhealing"] = "치유량/초과 치유량",
	["Healing Done Hit Details"] = "치유량 적중 상세정보",

	["Critical Heal Rate"] = "치유 극대화율",
	["Critical Heal Rate Details"] = "치유 극대화율 상세정보",

	["Healing Taken"] = "받은 치유량",
	["Fight Healing Taken"] = "최근 받은 치유량",

	["Healing"] = "치유량",
	["Healing Options."] = "치유량 설정입니다.",
	["Colors"] = "색상",
	["Color options"] = "색상 설정",
	["Healing"] = "치유량",
	["Overhealing"] = "초과 치유량",

	["Tick"] = "틱당",
	["Critical Heal"] = "치유 극대화",
	["Heal"] = "치유",

	["Heal Data"] = "치유 데이터",
	["Information about each heal used."] = "사용한 각 치유에 대한 정보입니다.",
	["Hit Data"] = "적중 데이터",
	["Detailed information about each attack used."] = "사용한 각 공격에 대한 상세정보입니다.",
} end)

-----------------------------------------------------------------------------

local AssessmentHealing = Assessment:NewModule("Healing", "AceEvent-2.0", "Parser-3.0")
AssessmentHealing.db = Assessment:AcquireDBNamespace("Healing")
Assessment:RegisterDefaults("Healing", "profile", {
	enable					= true,
	gatherHealDataFilter	= "ALL",
	gatherHitDataFilter		= "ALL",
	healColor				= { 1, 1, 1, 0 },
	overhealColor			= { 0, 0, 0, 0.5 },
})

local profile
local gatherHealDataMethod
local gatherHitDataMethod

local emptyMethod = Assessment.EmptyMethod
local new = Assessment.NewTable
local pairs = pairs

-----------------------------------------------------------------------------

Assessment.TagData.HealingDone =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local healingDone = fightData.healingDone
				if not healingDone then return 0 end

				local playerHealing = healingDone[data.key]
				if not playerHealing then return 0 end

				return playerHealing.amount
			 end,
	data = "healing",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.OverhealingDone =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local healingDone = fightData.healingDone
				if not healingDone then return 0 end
				
				local playerHealing = healingDone[data.key]
				if not playerHealing then return 0 end

				return playerHealing.overhealAmount
			 end,
	data = "healing",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.HealingTaken =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local healingTaken = fightData.healingTaken
				if not healingTaken then return 0 end
				
				local playerHealing = healingTaken[data.key]
				if not playerHealing then return 0 end

				return playerHealing.amount
			 end,
	data = "healing",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.OverhealingTaken =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local healingTaken = fightData.healingTaken
				if not healingTaken then return 0 end
				
				local playerHealing = healingTaken[data.key]
				if not playerHealing then return 0 end

				return playerHealing.overhealAmount
			 end,
	data = "healing",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.OverhealAmount =
{
	tableData = "value.overhealAmount",
	data = "healing",
	numeric = true,
}

-----------------------------------------------------------------------------

Assessment.ListType.EnumerateTable.updateData.healingDone = "healing"
Assessment.ListType.EnumerateTable.updateData.healingTaken = "healing"

-----------------------------------------------------------------------------

Assessment.ListType.Healers =
{
	method = function(fightData)
				if not fightData or not fightData.healingDone then return emptyMethod end
				return pairs(fightData.healingDone)
			 end,
	data = "healing",
	description = L["Healers List"],
}

local healIteratorText = L["Heal"]
local overhealIteratorText = L["Overheal"]
local function HealOverhealIterator(state, index)
	if not index then 
		return healIteratorText, state.amount-state.overhealAmount
	elseif index == healIteratorText then
		return overhealIteratorText, state.overhealAmount
	end
end

Assessment.ListType.HealOverheal =
{
	method = function(fightData, ...)
				local data = fightData.healingDone
				if not data then return emptyMethod end

				for i=1,select('#',...) do
					if not data then return emptyMethod end
					data = data[select(i, ...)]
				end
				if not data then return emptyMethod end
				
				return HealOverhealIterator, data, nil
			 end,

	data = "healing",
	description = L["Heal/Overheal"]
}

-----------------------------------------------------------------------------

local keyToColorNameMap = { [L["Heal"]] = "healColor", [L["Overheal"]] = "overhealColor" }
Assessment.ColorType.HealOverheal = 
{
	method = function(fightData, data)
				local colorName = keyToColorNameMap[data.key]
				if not colorName then return nil end
				return profile[colorName]
			 end,
	description = L["Heal/Overheal"]
}

-----------------------------------------------------------------------------

Assessment.DefaultDisplaySets.HealingDoneHitDetails =
{
	titleTag			= "[WindowParameter(1)]: [WindowParameter(2)]",
	listType			= "EnumerateTable", 
	listDataTag			= "healingDone,[WindowParameter(1)],healDetails,[WindowParameter(2)],hitDetails",
	barValueTag			= "Amount - OverhealAmount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { tag = "[Rank]. [Key] ([Count])" }, { tag = "[BarValue:Short]" } },
	description			= L["Healing Done Hit Details"],
	selectable			= false,
	defaultColor		= { 0.5, 0.8, 0.3 },
}

Assessment.DefaultDisplaySets.HealingDoneDetails =
{
	titleTag			= "[(FightIndex ~= 0):\"Fight \"]Healing Done: [WindowParameter(1)]",
	listType			= "EnumerateTable", 
	listDataTag			= "healingDone,[WindowParameter(1)],healDetails",
	barValueTag			= "Amount - OverhealAmount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { tag = "[Rank]. [Key] ([Count])" }, { tag = "[BarValue:Short]" } },
	description			= L["Healing Done Details"],
	selectable			= false,
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Healing Done Hit Details"],
	onClickData			= "[WindowParameter(1)],[Key]",
	defaultColor		= { 0.5, 0.8, 0.3 },
}

Assessment.DefaultDisplaySets.HealingDone =
{
	titleTag			= "Healing Done: [TotalBarValue:Short]",
	listType			= "GroupMembers", 
	barValueTag			= "HealingDone - OverhealingDone",
	filterTag			= "BarValue ~= 0",
	fightIndex			= 0,
	text				= { { }, { tag= "[BarValue:Short]" }, { tag = "[(BarValue*100/TotalBarValue):Format(\"%.1f%%\")]", visible = false }, count = 3 },
	description			= L["Healing Done"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Healing Done Details"],
	onClickData			= "[Key]",
	defaultColor		= { 0.5, 0.8, 0.3 },
	colorTypeList		= { "Class", count = 1 },
}

Assessment.DefaultDisplaySets.FightHealingDone =
{
	titleTag			= "Fight Healing Done: [TotalBarValue:Short]",
	listType			= "GroupMembers", 
	barValueTag			= "HealingDone - OverhealingDone", 
	filterTag			= "BarValue ~= 0",
	fightIndex			= 1,
	text				= { { }, { tag= "[BarValue:Short]" }, { tag = "[(BarValue*100/TotalBarValue):Format(\"%.1f%%\")]", visible = false }, count = 3 },
	description			= L["Fight Healing Done"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Healing Done Details"],
	onClickData			= "[Key]",
	defaultColor		= { 0.5, 0.8, 0.3 },
	colorTypeList		= { "Class", count = 1 },
}

Assessment.DefaultDisplaySets.HPS =
{
	titleTag			= "HPS: [CombatDuration:FormatDuration]",
	listType			= "GroupMembers", 
	barValueTag			= "(HealingDone - OverhealingDone) / Maximum(CombatDuration, 1)", 
	filterTag			= "BarValue ~= 0",
	fightIndex			= 1,
	text				= { { }, { tag= "[BarValue:Format(\"%.1f\")]" }, { tag = "[(BarValue*100/TotalBarValue):Format(\"%.1f%%\")]", visible = false }, count = 3 },
	description			= L["HPS"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Healing Done Details"],
	onClickData			= "[Key]",
	favorite			= false,
	defaultColor		= { 0.3, 0.8, 0.5 },
	colorTypeList		= { "Class", count = 1 },
}

Assessment.DefaultDisplaySets.HealingOverhealing =
{
	titleTag			= "[WindowParameter(1)]",
	listType			= "HealOverheal", 
	listDataTag			= "[WindowParameterList]",
	barValueTag			= "Value", 
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { }, { tag= "[BarValue:Short]" } },
	description			= L["Healing/Overhealing"],
	colorTypeList		= { "HealOverheal", count = 1 },
	sortKeyTag			= "Rank",
	descendingSort		= false,
	favorite			= false,
	selectable			= false,
}

Assessment.DefaultDisplaySets.RawHealingDoneHitDetails =
{
	titleTag			= "[WindowParameter(1)]: [WindowParameter(2)]",
	listType			= "EnumerateTable", 
	listDataTag			= "healingDone,[WindowParameter(1)],healDetails,[WindowParameter(2)],hitDetails",
	barValueTag			= "Amount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { tag = "[Rank]. [Key] ([Count])" }, { tag = "[BarValue:short][IsReport:\" Raw\"]" }, { tag = "[(Amount-OverhealAmount):Short] Actual, [OverhealAmount:Short] Overheal ([(OverhealAmount*100/Amount):Format(\"%.1f%%\")])", visible = false }, count = 3 },
	description			= L["Raw Healing Done Hit Details"],
	selectable			= false,
	segmentEnable		= true,
	segmentDisplaySet	= L["Healing/Overhealing"],
	segmentData			= "[WindowParameter(1)],healDetails,[WindowParameter(2)],hitDetails,[Key]",
	defaultColor		= { 0.3, 0.8, 0.8 },
}

Assessment.DefaultDisplaySets.RawHealingDoneDetails =
{
	titleTag			= "[(FightIndex ~= 0):\"Fight \"]Raw Healing Done: [WindowParameter(1)]",
	listType			= "EnumerateTable", 
	listDataTag			= "healingDone,[WindowParameter(1)],healDetails",
	barValueTag			= "Amount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { tag = "[Rank]. [Key] ([Count])" }, { tag = "[BarValue:Short][IsReport:\" Raw\"]" }, { tag = "[(Amount-OverhealAmount):Short] Actual, [OverhealAmount:Short] Overheal ([(OverhealAmount*100/Amount):Format(\"%.1f%%\")])", visible = false }, count = 3 },
	description			= L["Raw Healing Done Details"],
	selectable			= false,
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Raw Healing Done Hit Details"],
	onClickData			= "[WindowParameter(1)],[Key]",
	segmentEnable		= true,
	segmentDisplaySet	= L["Healing/Overhealing"],
	segmentData			= "[WindowParameter(1)],healDetails,[Key]",
	defaultColor		= { 0.3, 0.8, 0.8 },
}

Assessment.DefaultDisplaySets.FightRawHealingDone =
{
	titleTag			= "Fight Raw Healing Done: [TotalBarValue:Short]",
	listType			= "GroupMembers", 
	barValueTag			= "HealingDone", 
	filterTag			= "BarValue ~= 0",
	fightIndex			= 1,
	text				= { { }, { tag = "[BarValue:Short][IsReport:\" Raw\"]" }, { tag = "[(HealingDone-OverhealingDone):Short] Actual, [OverhealingDone:Short] Overheal ([(OverhealingDone*100/HealingDone):Format(\"%.1f%%\")])", visible = false }, count = 3 },
	description			= L["Fight Raw Healing Done"],
	favorite			= false,
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Raw Healing Done Details"],
	onClickData			= "[Key]",
	segmentEnable		= true,
	segmentDisplaySet	= L["Healing/Overhealing"],
	segmentData			= "[Key]",
	defaultColor		= { 0.3, 0.8, 0.8 },
	colorTypeList		= { "Class", count = 1 },
}

Assessment.DefaultDisplaySets.RawHealingDone =
{
	titleTag			= "Raw Healing: [TotalBarValue:Short]",
	listType			= "GroupMembers", 
	barValueTag			= "HealingDone", 
	filterTag			= "BarValue ~= 0",
	fightIndex			= 0,
	text				= { { }, { tag = "[BarValue:Short][IsReport:\" Raw\"]" }, { tag = "[(HealingDone-OverhealingDone):Short] Actual, [OverhealingDone:Short] Overheal ([(OverhealingDone*100/HealingDone):Format(\"%.1f%%\")])", visible = false }, count = 3 },
	description			= L["Raw Healing Done"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Raw Healing Done Details"],
	onClickData			= "[Key]",
	segmentEnable		= true,
	segmentDisplaySet	= L["Healing/Overhealing"],
	segmentData			= "[Key]",
	defaultColor		= { 0.3, 0.8, 0.8 },
	colorTypeList		= { "Class", count = 1 },
}

Assessment.DefaultDisplaySets.CriticalHealRateDetails =
{
	titleTag			= "Critical Heal Rate: [WindowParameter(1)]",
	listType			= "EnumerateTable", 
	listDataTag			= "healingDone,[windowParameter(1)],healDetails",
	barValueTag			= "GetValueField(\"hitDetails.Critical Heal.count\") / (GetValueField(\"hitDetails.Critical Heal.count\")+GetValueField(\"hitDetails.Heal.count\"))",
	filterTag			= "BarValue == BarValue", -- This is a NaN check
	fightIndex			= -1,
	text				= { { tag = "[Rank]. [Key] ([GetValueField(\"hitDetails.Critical Heal.count\")||0]/[(GetValueField(\"hitDetails.Critical Heal.count\")+GetValueField(\"hitDetails.Heal.count\"))])" }, 
							{ tag = "[(100*BarValue):Format(\"%.1f%%\")]" },
							count = 2,
						  },
	description			= L["Critical Heal Rate Details"],
	selectable			= false,
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Healing Done Hit Details"],
	onClickData			= "[WindowParameter(1)],[Key]",
	defaultColor		= { 0.8, 0.3, 0.8 },
}

Assessment.DefaultDisplaySets.CriticalHealRate =
{
	titleTag			= L["Critical Heal Rate"],
	listType			= "GroupMembers", 
	barValueTag			= "SumField(\"healingDone.\".Key.\".healDetails.*.hitDetails.Critical Heal.count\"):SetVariable(\"CriticalHealCount\") / (GetVariable(\"CriticalHealCount\")+SumField(\"healingDone.\".Key.\".healDetails.*.hitDetails.Heal.count\")):SetVariable(\"TotalHealCount\")", 
	filterTag			= "BarValue == BarValue", -- This is a NaN check
	fightIndex			= 0,
	text				= { { tag = "[Rank]. [Key] ([GetVariable(\"CriticalHealCount\")]/[GetVariable(\"TotalHealCount\")])" }, 
							{ tag= "[(BarValue*100):Format(\"%.1f%%\")]" }, 
							count = 2,
						  },
	description			= L["Critical Heal Rate"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Critical Heal Rate Details"],
	onClickData			= "[Key]",
	defaultColor		= { 0.8, 0.3, 0.8 },
	colorTypeList		= { "Class", count = 1 },
	favorite			= false,
}

-----------------------------------------------------------------------------

Assessment.Options.args.data.args.healing =
{
	type = "group",
	name = L["Healing"],
	desc = L["Healing Options."],
	args =
	{
		gatherHealData =
		{
			type = "text",
			name = L["Heal Data"],
			desc = L["Information about each heal used."],
			get = function() return profile.gatherHealDataFilter end,
			set = function(v) profile.gatherHealDataFilter = v AssessmentHealing:OnInitialize() end,
			validate = Assessment.GatherFilterValidate,
		},
		gatherHitData =
		{
			type = "text",
			name = L["Hit Data"],
			desc = L["Detailed information about each attack used."],
			get = function() return profile.gatherHitDataFilter end,
			set = function(v) profile.gatherHitDataFilter = v AssessmentHealing:OnInitialize() end,
			validate = Assessment.GatherFilterValidate,
		},
		color =
		{
			type = "group",
			name = L["Colors"],
			desc = L["Color options"],
			args =
			{
				healing =
				{
					type = "color",
					name = L["Healing"],
					desc = L["Healing"],
					get = function() return unpack(profile.healColor) end,
					set = function(r, g, b, a) profile.healColor = { r, g, b, a } Assessment:RedrawAll() end,
					hasAlpha = true
				},
				overHealing =
				{
					type = "color",
					name = L["Overhealing"],
					desc = L["Overhealing"],
					get = function() return unpack(profile.overhealColor) end,
					set = function(r, g, b, a) profile.overhealColor = { r, g, b, a } Assessment:RedrawAll() end,
					hasAlpha = true
				},
			},
		},
	}
}

-----------------------------------------------------------------------------

Assessment.DefaultTags.HealingDone =
{
	tag				= "[HealingDone:Short]",
	fightIndex		= 0,
	description		= L["Healing Done"],
}

Assessment.DefaultTags.FightHealingDone =
{
	tag				= "[HealingDone:Short]",
	fightIndex		= 1,
	description		= L["Fight Healing Done"],
}

Assessment.DefaultTags.HealingTaken =
{
	tag				= "[HealingTaken:Short]",
	fightIndex		= 0,
	description		= L["Healing Taken"],
}

Assessment.DefaultTags.FightHealingTaken =
{
	tag				= "[HealingTaken:Short]",
	fightIndex		= 1,
	description		= L["Fight Healing Taken"],
}

Assessment.DefaultTags.HPS =
{
	tag				= "[(HealingTaken / Maximum(CombatDuration, 1)):Format(\"%.1f\")]",
	fightIndex		= 1,
	description		= L["HPS"],
}

-----------------------------------------------------------------------------

function AssessmentHealing:OnInitialize()
	profile = self.db.profile

	gatherHealDataMethod	= Assessment.GatherFilterList[profile.gatherHealDataFilter].method
	gatherHitDataMethod		= Assessment.GatherFilterList[profile.gatherHitDataFilter].method
end

function AssessmentHealing:OnEnable()
	self:RegisterParserEvent({ eventType = "Heal" }, "OnParserHealEvent")
end

function AssessmentHealing:OnDisable()
end

-----------------------------------------------------------------------------

local function AccumulateData(data, eventData)
	data.count			= data.count + 1
	data.amount			= data.amount + eventData.amount
	data.overhealAmount = data.overhealAmount + eventData.overhealAmount
end

-----------------------------------------------------------------------------

local function NewDataTable()
	local data = new()
	data.count			= 0
	data.amount			= 0
	data.overhealAmount	= 0
	return data
end

-----------------------------------------------------------------------------

local function UpdateHealingData(dataTable, member, eventData, healType)
	local data = dataTable[member]
	if not data then
		data = NewDataTable()
		data.healDetails = new()
		dataTable[member] = data
	end
	AccumulateData(data, eventData)
	
	if gatherHealDataMethod(member) then
		local healDetailsParent = data.healDetails
		local healDetails = healDetailsParent[eventData.abilityName]
		if not healDetails then
			healDetails = NewDataTable()
			healDetails.hitDetails = new()
			healDetailsParent[eventData.abilityName] = healDetails
		end
		AccumulateData(healDetails, eventData)

		if gatherHitDataMethod(member) then
			local healTypeDetailsParent = healDetails.hitDetails
			local healTypeDetails = healTypeDetailsParent[healType]
			if not healTypeDetails then
				healTypeDetails = NewDataTable()
				healTypeDetailsParent[healType] = healTypeDetails
			end
			AccumulateData(healTypeDetails, eventData)
		end
	end
end

-----------------------------------------------------------------------------

local tickHealText		= L["Tick"]
local criticalHealText	= L["Critical Heal"]
local healText			= L["Heal"]

function AssessmentHealing:OnParserHealEvent(eventData)
	local healType = eventData.isHoT and tickHealText or 
					 eventData.isCrit and criticalHealText or 
					 healText

	local time = GetTime()
	for data in self:IterateUpdateDataTables() do
		data.dataUpdateTimes.healing = time
		local healingDone = data.healingDone
		if not healingDone then
			healingDone = new()
			data.healingDone = healingDone
		end

		local healingTaken = data.healingTaken
		if not healingTaken then
			healingTaken = new()
			data.healingTaken = healingTaken
		end

		UpdateHealingData(healingDone, eventData.sourceName, eventData, healType)
		UpdateHealingData(healingTaken, eventData.recipientName, eventData, healType)
	end
end

-----------------------------------------------------------------------------
